"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

from six import StringIO

import cfnlint.config  # pylint: disable=E0401
import cfnlint.core  # pylint: disable=E0401

LOGGER = logging.getLogger("cfnlint")


class TestCli(BaseTestCase):
    """Test CLI processing"""

    def setUp(self):
        """Setup"""
        # Disable all but critical logger output, as both 'test_template_not_found' and
        # 'test_template_not_found_directory' like to print out errors to the console.
        logging.disable(logging.CRITICAL)

    def tearDown(self):
        """Tear Down"""
        logging.disable(logging.NOTSET)
        for handler in LOGGER.handlers:
            LOGGER.removeHandler(handler)

    def test_template_not_found(self):
        """Test template not found"""

        filename = "test/fixtures/templates/good/core/not_found.yaml"

        (args, filenames, _) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--ignore-bad-template"]
        )
        (_, _, matches) = cfnlint.core.get_template_rules(filenames[0], args)

        self.assertEqual(len(matches), 0)

    def test_template_not_found_directory(self):
        """Test template not found"""

        filename = "test/fixtures/templates/good/core"

        (args, filenames, _) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--ignore-bad-template"]
        )
        (_, _, matches) = cfnlint.core.get_template_rules(filenames[0], args)

        self.assertEqual(len(matches), 0)

    def test_template_invalid_yaml(self):
        """Test template not found"""
        filename = "test/fixtures/templates/bad/core/config_invalid_yaml.yaml"

        (args, filenames, _) = cfnlint.core.get_args_filenames(["--template", filename])
        (_, _, matches) = cfnlint.core.get_template_rules(filenames[0], args)

        self.assertEqual(len(matches), 1)

    def test_template_invalid_json(self):
        """Test template not found"""
        filename = "test/fixtures/templates/bad/core/config_invalid_json.json"

        (args, filenames, _) = cfnlint.core.get_args_filenames(["--template", filename])
        (_, _, matches) = cfnlint.core.get_template_rules(filenames[0], args)

        self.assertEqual(len(matches), 1)

    def test_template_invalid_yaml_ignore(self):
        """Test template not found"""
        filename = "test/fixtures/templates/bad/core/config_invalid_yaml.yaml"

        (args, filenames, _) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--ignore-bad-template"]
        )
        (_, _, matches) = cfnlint.core.get_template_rules(filenames[0], args)

        self.assertEqual(len(matches), 0)

    def test_template_invalid_yaml_ignore_by_rule_id(self):
        """Test template not found"""
        filename = "test/fixtures/templates/bad/core/config_invalid_yaml.yaml"

        (args, filenames, _) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--ignore-checks", "E0000"]
        )
        (_, _, matches) = cfnlint.core.get_template_rules(filenames[0], args)

        self.assertEqual(len(matches), 0)

    def test_template_invalid_json_ignore(self):
        """Test template not found"""
        filename = "test/fixtures/templates/bad/core/config_invalid_json.json"

        (args, filenames, _) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--ignore-bad-template"]
        )
        (_, _, matches) = cfnlint.core.get_template_rules(filenames[0], args)

        self.assertEqual(len(matches), 0)

    def test_template_via_stdin(self):
        """Test getting the template from stdin doesn't crash"""
        filename = "test/fixtures/templates/good/generic.yaml"
        with open(filename, "r") as fp:
            file_content = fp.read()

        with patch("sys.stdin", StringIO(file_content)):
            (_, filenames, _) = cfnlint.core.get_args_filenames([])
            assert filenames == [None]

        with patch("sys.stdin", StringIO(file_content)):
            (_, filenames, _) = cfnlint.core.get_args_filenames(
                ["--template", filename]
            )
            assert filenames == [filename]

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_template_config(self, yaml_mock):
        """Test template config"""
        yaml_mock.side_effect = [{}, {}]

        filename = "test/fixtures/templates/good/core/config_parameters.yaml"
        (
            args,
            _,
            _,
        ) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--ignore-bad-template"]
        )

        self.assertEqual(args.append_rules, [])
        self.assertEqual(args.append_rules, [])
        self.assertEqual(args.format, None)
        self.assertEqual(args.ignore_bad_template, True)
        self.assertEqual(args.ignore_checks, [])
        self.assertEqual(args.include_checks, ["W", "E"])
        self.assertEqual(args.listrules, False)
        self.assertEqual(args.debug, False)
        self.assertEqual(args.override_spec, None)
        self.assertEqual(args.custom_rules, None)
        self.assertEqual(args.regions, ["us-east-1"])
        self.assertEqual(
            args.templates, ["test/fixtures/templates/good/core/config_parameters.yaml"]
        )
        self.assertEqual(args.update_documentation, False)
        self.assertEqual(args.update_specs, False)
        self.assertEqual(args.output_file, None)

    def test_output_file(self):
        filename = "test/fixtures/templates/good/core/config_parameters.yaml"
        (
            args,
            _,
            _,
        ) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--output-file", "test_output.txt"]
        )

        self.assertEqual(args.output_file, "test_output.txt")

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_positional_template_parameters(self, yaml_mock):
        """Test overriding parameters"""
        yaml_mock.side_effect = [{}, {}]

        filename = "test/fixtures/templates/good/core/config_parameters.yaml"
        (args, _, _) = cfnlint.core.get_args_filenames(
            [filename, "--ignore-bad-template", "--ignore-checks", "E0000"]
        )

        self.assertEqual(args.append_rules, [])
        self.assertEqual(args.format, None)
        self.assertEqual(args.ignore_bad_template, True)
        self.assertEqual(args.ignore_checks, ["E0000"])
        self.assertEqual(args.include_checks, ["W", "E"])
        self.assertEqual(args.listrules, False)
        self.assertEqual(args.debug, False)
        self.assertEqual(args.override_spec, None)
        self.assertEqual(args.custom_rules, None)
        self.assertEqual(args.regions, ["us-east-1"])
        self.assertEqual(
            args.templates, ["test/fixtures/templates/good/core/config_parameters.yaml"]
        )
        self.assertEqual(args.update_documentation, False)
        self.assertEqual(args.update_specs, False)

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_override_parameters(self, yaml_mock):
        """Test overriding parameters"""
        yaml_mock.side_effect = [{}, {}]

        filename = "test/fixtures/templates/good/core/config_parameters.yaml"
        (args, _, _) = cfnlint.core.get_args_filenames(
            [
                "--template",
                filename,
                "--ignore-bad-template",
                "--ignore-checks",
                "E0000",
            ]
        )

        self.assertEqual(args.append_rules, [])
        self.assertEqual(args.format, None)
        self.assertEqual(args.ignore_bad_template, True)
        self.assertEqual(args.ignore_checks, ["E0000"])
        self.assertEqual(args.include_checks, ["W", "E"])
        self.assertEqual(args.listrules, False)
        self.assertEqual(args.debug, False)
        self.assertEqual(args.override_spec, None)
        self.assertEqual(args.custom_rules, None)
        self.assertEqual(args.regions, ["us-east-1"])
        self.assertEqual(
            args.templates, ["test/fixtures/templates/good/core/config_parameters.yaml"]
        )
        self.assertEqual(args.update_documentation, False)
        self.assertEqual(args.update_specs, False)

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_bad_config(self, yaml_mock):
        """Test bad formatting in config"""
        yaml_mock.side_effect = [{}, {}]

        filename = "test/fixtures/templates/bad/core/config_parameters.yaml"
        (args, _, _) = cfnlint.core.get_args_filenames(
            ["--template", filename, "--ignore-bad-template"]
        )

        self.assertEqual(args.append_rules, [])
        self.assertEqual(args.format, None)
        self.assertEqual(args.ignore_bad_template, True)
        self.assertEqual(args.ignore_checks, [])
        self.assertEqual(args.include_checks, ["W", "E"])
        self.assertEqual(args.listrules, False)
        self.assertEqual(args.debug, False)
        self.assertEqual(args.override_spec, None)
        self.assertEqual(args.custom_rules, None)
        self.assertEqual(args.regions, ["us-east-1"])
        self.assertEqual(args.templates, [filename])
        self.assertEqual(args.update_documentation, False)
        self.assertEqual(args.update_specs, False)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.testlib.testcase import BaseTestCase

import cfnlint.decode.cfn_json  # pylint: disable=E0401
import cfnlint.decode.cfn_yaml  # pylint: disable=E0401
from cfnlint.core import DEFAULT_RULESDIR  # pylint: disable=E0401
from cfnlint.rules import RulesCollection
from cfnlint.runner import Runner


class TestRunner(BaseTestCase):
    """Test Duplicates Parsing"""

    def setUp(self):
        """SetUp template object"""
        self.collection = RulesCollection(
            include_rules=["I"], include_experimental=True
        )
        rulesdirs = [DEFAULT_RULESDIR]
        for rulesdir in rulesdirs:
            self.collection.create_from_directory(rulesdir)
        (self.template, _) = cfnlint.decode.decode_str(
            """
            AWSTemplateFormatVersion: "2010-09-09"
            Description: >
                Template with all error levels: Warning, Error and Informational
            Resources:
                myTable:
                    Metadata:
                        cfn-lint:
                            config:
                                ignore_checks:
                                - I3011
                                - W1020
                    Type: "AWS::DynamoDB::Table"
                    Properties:
                        TableName: !Sub "TableName"
                        AttributeDefinitions:
                            - AttributeName: "Id"
                              AttributeType: "S" # Valid AllowedValue
                        KeySchema:
                            - AttributeName: "Id"
                              KeyType: "HASH"
                        ProvisionedThroughput:
                            ReadCapacityUnits: 5
                            WriteCapacityUnits: "5"
        """
        )

    def test_runner(self):
        """Success test"""
        runner = Runner(self.collection, "", self.template, ["us-east-1"], [])
        runner.transform()
        failures = runner.run()
        assert [] == failures, "Got failures {}".format(failures)

    def test_runner_mandatory_rules(self):
        """Success test"""
        runner = Runner(
            self.collection, "", self.template, ["us-east-1"], mandatory_rules=["W1020"]
        )
        runner.transform()
        failures = runner.run()
        self.assertEqual(len(failures), 1)

        runner = Runner(
            self.collection, "", self.template, ["us-east-1"], mandatory_rules=["W9000"]
        )
        runner.transform()
        failures = runner.run()
        self.assertEqual(len(failures), 0)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.testlib.testcase import BaseTestCase

import cfnlint.decode.node  # pylint: disable=E0401


class TestNodeSub(BaseTestCase):
    def test_success_init(self):
        """Test Dict Object"""
        template = cfnlint.decode.node.sub_node(
            {"Fn::Sub": ["${var}", {"var": {"Ref": "AWS::Region"}}]}, (0, 1), (2, 3)
        )

        self.assertTrue(template.is_valid())
        self.assertEqual(template.get_string_vars(), set(["var"]))
        self.assertEqual(template.get_defined_vars(), {"var": {"Ref": "AWS::Region"}})
        self.assertEqual(template.get_string(), "${var}")

        template = cfnlint.decode.node.sub_node(
            {"Fn::Sub": "${AWS::Region}"}, (0, 1), (2, 3)
        )

        self.assertTrue(template.is_valid())
        self.assertEqual(template.get_string_vars(), set(["AWS::Region"]))
        self.assertEqual(template.get_defined_vars(), {})
        self.assertEqual(template.get_string(), "${AWS::Region}")

    def test_fail_init(self):
        """All the ways I can think of configuring Fn::Sub wrong"""

        # extra element in array
        template = cfnlint.decode.node.sub_node(
            {"Fn::Sub": ["{var}", {"var": {"Ref": "AWS::Region"}}, "extra element"]},
            (0, 1),
            (2, 3),
        )

        self.assertFalse(template.is_valid())

        # first element is a dict and not a string
        template = cfnlint.decode.node.sub_node(
            {
                "Fn::Sub": [
                    {"var": {"Ref": "AWS::Region"}},
                    "string",
                ]
            },
            (0, 1),
            (2, 3),
        )

        self.assertFalse(template.is_valid())

        # second element is a string not a dict
        template = cfnlint.decode.node.sub_node(
            {
                "Fn::Sub": [
                    "string",
                    "string",
                ]
            },
            (0, 1),
            (2, 3),
        )

        self.assertFalse(template.is_valid())

        # multiple elements at the Fn::Sub level
        template = cfnlint.decode.node.sub_node(
            {
                "Fn::Sub": [],
                "Fn::Join": [],
            },
            (0, 1),
            (2, 3),
        )

        self.assertFalse(template.is_valid())

        template = cfnlint.decode.node.sub_node(
            {
                "Fn::Join": [],
            },
            (0, 1),
            (2, 3),
        )

        self.assertFalse(template.is_valid())
        self.assertEqual(template.get_string_vars(), set())
        self.assertEqual(template.get_defined_vars(), {})
        self.assertEqual(template.get_string(), "")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.properties.UnwantedBasedOnValue import (
    UnwantedBasedOnValue,  # pylint: disable=E0401
)


class TestUnwantedBasedOnValue(BaseRuleTestCase):
    def setUp(self):
        """Setup"""
        super(TestUnwantedBasedOnValue, self).setUp()
        self.collection.register(UnwantedBasedOnValue())
        self.success_templates = [
            "test/fixtures/templates/good/resources/properties/unwanted_based_on_value.yaml"
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/properties/unwanted_based_on_value.yaml",
            3,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.functions.Join import Join  # pylint: disable=E0401


class TestRulesJoin(BaseRuleTestCase):
    """Test Rules Get Att"""

    def setUp(self):
        """Setup"""
        super(TestRulesJoin, self).setUp()
        self.collection.register(Join())
        self.success_templates = [
            "test/fixtures/templates/good/functions/join.yaml",
        ]

    def test_functions(self):
        """Test some of the base functions"""
        rule = Join()

        output = rule._normalize_getatt("Resource.Output")
        self.assertEqual(output[0], "Resource")
        self.assertEqual(output[1], "Output")

        output = rule._normalize_getatt("Resource.Outputs.Output")
        self.assertEqual(output[0], "Resource")
        self.assertEqual(output[1], "Outputs.Output")

        output = rule._normalize_getatt(["Resource", "Output"])
        self.assertEqual(output[0], "Resource")
        self.assertEqual(output[1], "Output")

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative("test/fixtures/templates/bad/functions/join.yaml", 11)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.testlib.testcase import BaseTestCase

from cfnlint.rules import CloudFormationLintRule  # pylint: disable=E0401


class TestCloudFormationRule(BaseTestCase):
    """Test CloudFormation Rule"""

    def test_base(self):
        """Test Base Rule"""

        class TestRule(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-python-lint/"
            tags = ["resources"]

        self.assertEqual(TestRule.id, "E1000")
        self.assertEqual(TestRule.shortdesc, "Test Rule")
        self.assertEqual(TestRule.description, "Test Rule Description")
        self.assertEqual(
            TestRule.source_url,
            "https://github.com/aws-cloudformation/cfn-python-lint/",
        )
        self.assertEqual(TestRule.tags, ["resources"])

    def test_config(self):
        """Test Configuration"""

        class TestRule(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule"
            source_url = "https://github.com/aws-cloudformation/cfn-python-lint/"
            tags = ["resources"]

            def __init__(self):
                """Init"""
                super(TestRule, self).__init__()
                self.config_definition = {
                    "testBoolean": {"default": True, "type": "boolean"},
                    "testString": {"default": "default", "type": "string"},
                    "testInteger": {"default": 1, "type": "integer"},
                }
                self.configure()

            def get_config(self):
                """Get the Config"""
                return self.config

        rule = TestRule()
        config = rule.get_config()
        self.assertTrue(config.get("testBoolean"))
        self.assertEqual(config.get("testString"), "default")
        self.assertEqual(config.get("testInteger"), 1)

        rule.configure({"testBoolean": "false", "testString": "new", "testInteger": 2})

        self.assertFalse(config.get("testBoolean"))
        self.assertEqual(config.get("testString"), "new")
        self.assertEqual(config.get("testInteger"), 2)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.elb.Elb import Elb  # pylint: disable=E0401


class TestPropertyElb(BaseRuleTestCase):
    """Test Elb Property Configuration"""

    def setUp(self):
        """Setup"""
        super(TestPropertyElb, self).setUp()
        self.collection.register(Elb())
        self.success_templates = [
            "test/fixtures/templates/good/resources/elb/properties.yaml"
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/elb/properties.yaml", 6
        )

    def test_alb_subnets(self):
        """Test ALB Subnet Logic"""
        rule = Elb()

        # Failure when 1 subnet defined
        props = {"Subnets": ["subnet-123456"]}

        matches = rule.check_alb_subnets(props, ["Resources", "ALB", "Properties"], {})
        self.assertEqual(len(matches), 1)

        # No Failure when 2 subnets defined
        props = {"Type": "application", "Subnets": ["subnet-123456", "subnet-abcdef"]}

        matches = rule.check_alb_subnets(props, ["Resources", "ALB", "Properties"], {})
        self.assertEqual(len(matches), 0)

        # Failure when 1 SubnetMapping defined
        props = {
            "Type": "application",
            "SubnetMappings": [{"SubnetId": "subnet-123456"}],
        }

        matches = rule.check_alb_subnets(props, ["Resources", "ALB", "Properties"], {})
        self.assertEqual(len(matches), 1)

        # No Failure when 2 SubnetMapping defined
        props = {
            "Type": "application",
            "SubnetMappings": [
                {"SubnetId": "subnet-123456"},
                {"SubnetId": "subnet-abcdef"},
            ],
        }

        matches = rule.check_alb_subnets(props, ["Resources", "ALB", "Properties"], {})
        self.assertEqual(len(matches), 0)

        # No Failure when 1 Subnet and NLB
        props = {"Type": "network", "Subnets": ["subnet-123456"]}

        matches = rule.check_alb_subnets(props, ["Resources", "NLB", "Properties"], {})
        self.assertEqual(len(matches), 0)

    def test_loadbalancer_attributes(self):
        """Test LoadBalancer Attributes logic"""
        rule = Elb()

        props = {"Type": "network"}

        elb_type = rule.get_loadbalancer_type(props)
        self.assertEqual(elb_type, "network")

        props = {"Type": "application"}

        elb_type = rule.get_loadbalancer_type(props)
        self.assertEqual(elb_type, "application")

        props = {}

        elb_type = rule.get_loadbalancer_type(props)
        self.assertEqual(elb_type, "application")

        props = {"Type": {"Ref": "LoadBalancerType"}}

        elb_type = rule.get_loadbalancer_type(props)
        self.assertEqual(elb_type, None)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.functions.Length import Length


class TestRulesLength(BaseRuleTestCase):
    def setUp(self):
        super(TestRulesLength, self).setUp()
        self.collection.register(Length())
        self.success_templates = [
            "test/fixtures/templates/good/functions/length.yaml",
        ]

    def test_file_positive(self):
        self.helper_file_positive()

    def test_file_negative_missing_transform(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/functions/lengthWithoutTransform.yaml", 1
        )

    def test_file_negative_wrong_types(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/functions/lengthWrongTypes.yaml", 7
        )
